package com.sun.xml.internal.bind.v2.model.impl;

import javax.xml.bind.annotation.XmlSchemaType;
import javax.xml.bind.annotation.XmlList;
import javax.xml.bind.annotation.XmlElementWrapper;
import javax.xml.bind.annotation.XmlAttachmentRef;
import javax.xml.bind.annotation.XmlMimeType;
import javax.xml.bind.annotation.XmlInlineBinaryData;
import javax.xml.bind.annotation.XmlIDREF;
import javax.xml.bind.annotation.XmlID;
import com.sun.xml.internal.bind.v2.util.EditDistance;
import java.util.AbstractList;
import java.util.HashSet;
import java.util.Comparator;
import java.lang.reflect.Method;
import com.sun.xml.internal.bind.v2.runtime.Location;
import com.sun.xml.internal.bind.v2.model.core.PropertyKind;
import java.util.Collection;
import java.util.ArrayList;
import java.util.Set;
import com.sun.xml.internal.bind.v2.model.annotation.MethodLocatable;
import java.util.TreeSet;
import java.util.LinkedHashMap;
import javax.xml.bind.annotation.adapters.XmlJavaTypeAdapter;
import java.util.Map;
import javax.xml.bind.annotation.XmlMixed;
import javax.xml.bind.annotation.XmlAnyElement;
import javax.xml.bind.annotation.XmlElementRefs;
import javax.xml.bind.annotation.XmlElementRef;
import javax.xml.bind.annotation.XmlElements;
import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlValue;
import javax.xml.bind.annotation.XmlAttribute;
import javax.xml.bind.annotation.XmlAnyAttribute;
import javax.xml.bind.annotation.XmlTransient;
import javax.xml.bind.annotation.XmlAccessorType;
import com.sun.xml.internal.bind.v2.model.core.ValuePropertyInfo;
import com.sun.xml.internal.bind.annotation.OverrideAnnotationOf;
import javax.xml.bind.annotation.XmlAccessType;
import java.util.Collections;
import java.util.List;
import java.util.Iterator;
import com.sun.xml.internal.bind.v2.model.core.ID;
import com.sun.xml.internal.bind.v2.model.core.PropertyInfo;
import com.sun.xml.internal.bind.v2.model.core.NonElement;
import com.sun.xml.internal.bind.v2.runtime.IllegalAnnotationException;
import javax.xml.bind.annotation.XmlAccessOrder;
import javax.xml.bind.annotation.XmlAccessorOrder;
import javax.xml.bind.annotation.XmlType;
import com.sun.xml.internal.bind.v2.model.annotation.Locatable;
import java.util.HashMap;
import java.lang.annotation.Annotation;
import com.sun.istack.internal.FinalArrayList;
import javax.xml.namespace.QName;
import com.sun.xml.internal.bind.v2.model.core.Element;
import com.sun.xml.internal.bind.v2.model.core.ClassInfo;

public class ClassInfoImpl<T, C, F, M> extends TypeInfoImpl<T, C, F, M> implements ClassInfo<T, C>, Element<T, C>
{
    protected final C clazz;
    private final QName elementName;
    private final QName typeName;
    private FinalArrayList<PropertyInfoImpl<T, C, F, M>> properties;
    private String[] propOrder;
    private ClassInfoImpl<T, C, F, M> baseClass;
    private boolean baseClassComputed;
    private boolean hasSubClasses;
    protected PropertySeed<T, C, F, M> attributeWildcard;
    private M factoryMethod;
    private static final ClassInfoImpl.SecondaryAnnotation[] SECONDARY_ANNOTATIONS;
    private static final Annotation[] EMPTY_ANNOTATIONS;
    private static final HashMap<Class, Integer> ANNOTATION_NUMBER_MAP;
    private static final String[] DEFAULT_ORDER;
    
    ClassInfoImpl(ModelBuilder<T, C, F, M> builder, Locatable upstream, C clazz) {
        super(builder, upstream);
        this.baseClassComputed = false;
        this.hasSubClasses = false;
        this.factoryMethod = null;
        this.clazz = clazz;
        assert clazz != null;
        this.elementName = this.parseElementName(clazz);
        XmlType t = (XmlType)this.reader().getClassAnnotation(XmlType.class, clazz, this);
        this.typeName = this.parseTypeName(clazz, t);
        if (t != null) {
            String[] propOrder = t.propOrder();
            if (propOrder.length == 0) {
                this.propOrder = null;
            }
            else if (propOrder[0].length() == 0) {
                this.propOrder = ClassInfoImpl.DEFAULT_ORDER;
            }
            else {
                this.propOrder = propOrder;
            }
        }
        else {
            this.propOrder = ClassInfoImpl.DEFAULT_ORDER;
        }
        XmlAccessorOrder xao = (XmlAccessorOrder)this.reader().getPackageAnnotation(XmlAccessorOrder.class, clazz, this);
        if (xao != null && xao.value() == XmlAccessOrder.UNDEFINED) {
            this.propOrder = null;
        }
        xao = (XmlAccessorOrder)this.reader().getClassAnnotation(XmlAccessorOrder.class, clazz, this);
        if (xao != null && xao.value() == XmlAccessOrder.UNDEFINED) {
            this.propOrder = null;
        }
        if (this.nav().isInterface(clazz)) {
            builder.reportError(new IllegalAnnotationException(Messages.CANT_HANDLE_INTERFACE.format(new Object[] { this.nav().getClassName(clazz) }), this));
        }
        if (!this.hasFactoryConstructor(t) && !this.nav().hasDefaultConstructor(clazz)) {
            if (this.nav().isInnerClass(clazz)) {
                builder.reportError(new IllegalAnnotationException(Messages.CANT_HANDLE_INNER_CLASS.format(new Object[] { this.nav().getClassName(clazz) }), this));
            }
            else if (this.elementName != null) {
                builder.reportError(new IllegalAnnotationException(Messages.NO_DEFAULT_CONSTRUCTOR.format(new Object[] { this.nav().getClassName(clazz) }), this));
            }
        }
        return;
    }
    
    public ClassInfoImpl<T, C, F, M> getBaseClass() {
        if (!this.baseClassComputed) {
            C s = this.nav().getSuperClass(this.clazz);
            if (s == null || s == this.nav().asDecl(Object.class)) {
                this.baseClass = null;
            }
            else {
                NonElement<T, C> b = this.builder.getClassInfo(s, true, this);
                if ((b instanceof ClassInfoImpl)) {
                    this.baseClass = (ClassInfoImpl)b;
                    this.baseClass.hasSubClasses = true;
                }
                else {
                    this.baseClass = null;
                }
            }
            this.baseClassComputed = true;
        }
        return this.baseClass;
    }
    
    public final Element<T, C> getSubstitutionHead() {
        ClassInfoImpl<T, C, F, M> c;
        for (c = this.getBaseClass(); c != null && !c.isElement(); c = c.getBaseClass()) {}
        return c;
    }
    
    public final C getClazz() {
        return this.clazz;
    }
    
    @Deprecated
    public ClassInfoImpl<T, C, F, M> getScope() {
        return null;
    }
    
    public final T getType() {
        return this.nav().use(this.clazz);
    }
    
    public boolean canBeReferencedByIDREF() {
        for (PropertyInfo<T, C> p : this.getProperties()) {
            if (p.id() == ID.ID) {
                return true;
            }
            continue;
        }
        ClassInfoImpl<T, C, F, M> base = this.getBaseClass();
        return base != null && base.canBeReferencedByIDREF();
    }
    
    public final String getName() {
        return this.nav().getClassName(this.clazz);
    }
    
    public <A extends Annotation> A readAnnotation(Class<A> a) {
        return this.reader().getClassAnnotation(a, this.clazz, this);
    }
    
    public Element<T, C> asElement() {
        if (this.isElement()) {
            return this;
        }
        return null;
    }
    
    public List<? extends PropertyInfo<T, C>> getProperties() {
        if (this.properties != null) {
            return this.properties;
        }
        XmlAccessType at = this.getAccessType();
        this.properties = new FinalArrayList();
        this.findFieldProperties(this.clazz, at);
        this.findGetterSetterProperties(at);
        if (this.propOrder == ClassInfoImpl.DEFAULT_ORDER || this.propOrder == null) {
            XmlAccessOrder ao = this.getAccessorOrder();
            if (ao == XmlAccessOrder.ALPHABETICAL) {
                Collections.sort(this.properties);
            }
        }
        else {
            ClassInfoImpl.PropertySorter sorter = new ClassInfoImpl.PropertySorter();
            for (PropertyInfoImpl p : this.properties) {
                ClassInfoImpl.PropertySorter.access$000(sorter, p);
            }
            Collections.sort(this.properties, sorter);
            sorter.checkUnusedProperties();
        }
        PropertyInfoImpl<Object, Object, Object, Object> vp = null;
        PropertyInfoImpl<Object, Object, Object, Object> ep = null;
        for (PropertyInfoImpl p2 : this.properties) {
            switch (p2.kind()) {
                case ELEMENT:
                case REFERENCE:
                case MAP: {
                    ep = p2;
                    continue;
                }
                case VALUE: {
                    if (vp != null) {
                        this.builder.reportError(new IllegalAnnotationException(Messages.MULTIPLE_VALUE_PROPERTY.format(new Object[0]), vp, p2));
                    }
                    if (this.getBaseClass() != null) {
                        this.builder.reportError(new IllegalAnnotationException(Messages.XMLVALUE_IN_DERIVED_TYPE.format(new Object[0]), p2));
                    }
                    vp = p2;
                    break;
                }
                case ATTRIBUTE: {
                    continue;
                }
                default: {
                    assert false;
                    continue;
                }
            }
        }
        if (ep != null && vp != null) {
            this.builder.reportError(new IllegalAnnotationException(Messages.ELEMENT_AND_VALUE_PROPERTY.format(new Object[0]), vp, ep));
        }
        return this.properties;
    }
    
    private void findFieldProperties(C c, XmlAccessType at) {
        C sc = this.nav().getSuperClass(c);
        if (this.shouldRecurseSuperClass(sc)) {
            this.findFieldProperties(sc, at);
        }
        for (F f : this.nav().getDeclaredFields(c)) {
            Annotation[] annotations = this.reader().getAllFieldAnnotations(f, this);
            boolean isDummy = this.reader().hasFieldAnnotation(OverrideAnnotationOf.class, f);
            if (this.nav().isTransient(f)) {
                if (hasJAXBAnnotation(annotations)) {
                    this.builder.reportError(new IllegalAnnotationException(Messages.TRANSIENT_FIELD_NOT_BINDABLE.format(new Object[] { this.nav().getFieldName(f) }), getSomeJAXBAnnotation(annotations)));
                }
                else {
                    continue;
                }
            }
            else if (this.nav().isStaticField(f)) {
                if (hasJAXBAnnotation(annotations)) {
                    this.addProperty(this.createFieldSeed(f), annotations, false);
                }
                else {
                    continue;
                }
            }
            else {
                if (at == XmlAccessType.FIELD || (at == XmlAccessType.PUBLIC_MEMBER && this.nav().isPublicField(f)) || hasJAXBAnnotation(annotations)) {
                    if (isDummy) {
                        ClassInfo<T, C> top;
                        for (top = this.getBaseClass(); top != null && top.getProperty("content") == null; top = top.getBaseClass()) {}
                        DummyPropertyInfo<Object, Object, Object, Object> prop = (DummyPropertyInfo)top.getProperty("content");
                        PropertySeed<T, C, F, M> seed = this.createFieldSeed(f);
                        prop.addType(this.createReferenceProperty(seed));
                    }
                    else {
                        this.addProperty(this.createFieldSeed(f), annotations, false);
                    }
                }
                this.checkFieldXmlLocation(f);
            }
        }
    }
    
    public final boolean hasValueProperty() {
        ClassInfoImpl<T, C, F, M> bc = this.getBaseClass();
        if (bc != null && bc.hasValueProperty()) {
            return true;
        }
        for (PropertyInfo p : this.getProperties()) {
            if ((p instanceof ValuePropertyInfo)) {
                return true;
            }
            continue;
        }
        return false;
    }
    
    public PropertyInfo<T, C> getProperty(String name) {
        for (PropertyInfo<T, C> p : this.getProperties()) {
            if (p.getName().equals(name)) {
                return p;
            }
            continue;
        }
        return null;
    }
    
    protected void checkFieldXmlLocation(F f) {
    }
    
    private <T extends Annotation> T getClassOrPackageAnnotation(Class<T> type) {
        T t = this.reader().getClassAnnotation(type, this.clazz, this);
        if (t != null) {
            return t;
        }
        return this.reader().getPackageAnnotation(type, this.clazz, this);
    }
    
    private XmlAccessType getAccessType() {
        XmlAccessorType xat = (XmlAccessorType)this.getClassOrPackageAnnotation(XmlAccessorType.class);
        if (xat != null) {
            return xat.value();
        }
        return XmlAccessType.PUBLIC_MEMBER;
    }
    
    private XmlAccessOrder getAccessorOrder() {
        XmlAccessorOrder xao = (XmlAccessorOrder)this.getClassOrPackageAnnotation(XmlAccessorOrder.class);
        if (xao != null) {
            return xao.value();
        }
        return XmlAccessOrder.UNDEFINED;
    }
    
    public boolean hasProperties() {
        return !this.properties.isEmpty();
    }
    
    private static <T> T pickOne(T... args) {
        int length = args.length;
        int i = 0;
        while (i < length) {
            T arg = args[i];
            if (arg != null) {
                return arg;
            }
            ++i;
        }
        return null;
    }
    
    private static <T> List<T> makeSet(T... args) {
        List<T> l = new FinalArrayList();
        for (T arg : args) {
            if (arg != null) {
                l.add(arg);
            }
        }
        return l;
    }
    
    private void checkConflict(Annotation a, Annotation b) throws ClassInfoImpl.DuplicateException {
        assert b != null;
        if (a != null) {
            throw new DuplicateException(a, b);
        }
        return;
    }
    
    private void addProperty(PropertySeed<T, C, F, M> seed, Annotation[] annotations, boolean dummy) {
        XmlTransient t = null;
        XmlAnyAttribute aa = null;
        XmlAttribute a = null;
        XmlValue v = null;
        XmlElement e1 = null;
        XmlElements e2 = null;
        XmlElementRef r1 = null;
        XmlElementRefs r2 = null;
        XmlAnyElement xae = null;
        XmlMixed mx = null;
        OverrideAnnotationOf ov = null;
        int secondaryAnnotations = 0;
        try {
            for (Annotation ann : annotations) {
                Integer index = (Integer)ClassInfoImpl.ANNOTATION_NUMBER_MAP.get(ann.annotationType());
                if (index != null) {
                    switch ((int)index) {
                        case 0: {
                            this.checkConflict(t, ann);
                            t = (XmlTransient)ann;
                            break;
                        }
                        case 1: {
                            this.checkConflict(aa, ann);
                            aa = (XmlAnyAttribute)ann;
                            break;
                        }
                        case 2: {
                            this.checkConflict(a, ann);
                            a = (XmlAttribute)ann;
                            break;
                        }
                        case 3: {
                            this.checkConflict(v, ann);
                            v = (XmlValue)ann;
                            break;
                        }
                        case 4: {
                            this.checkConflict(e1, ann);
                            e1 = (XmlElement)ann;
                            break;
                        }
                        case 5: {
                            this.checkConflict(e2, ann);
                            e2 = (XmlElements)ann;
                            break;
                        }
                        case 6: {
                            this.checkConflict(r1, ann);
                            r1 = (XmlElementRef)ann;
                            break;
                        }
                        case 7: {
                            this.checkConflict(r2, ann);
                            r2 = (XmlElementRefs)ann;
                            break;
                        }
                        case 8: {
                            this.checkConflict(xae, ann);
                            xae = (XmlAnyElement)ann;
                            break;
                        }
                        case 9: {
                            this.checkConflict(mx, ann);
                            mx = (XmlMixed)ann;
                            break;
                        }
                        case 10: {
                            this.checkConflict(ov, ann);
                            ov = (OverrideAnnotationOf)ann;
                            break;
                        }
                        default: {
                            secondaryAnnotations |= 1 << (int)index - 20;
                            break;
                        }
                    }
                }
            }
            ClassInfoImpl.PropertyGroup group = null;
            int groupCount = 0;
            if (t != null) {
                group = ClassInfoImpl.PropertyGroup.TRANSIENT;
                ++groupCount;
            }
            if (aa != null) {
                group = ClassInfoImpl.PropertyGroup.ANY_ATTRIBUTE;
                ++groupCount;
            }
            if (a != null) {
                group = ClassInfoImpl.PropertyGroup.ATTRIBUTE;
                ++groupCount;
            }
            if (v != null) {
                group = ClassInfoImpl.PropertyGroup.VALUE;
                ++groupCount;
            }
            if (e1 != null || e2 != null) {
                group = ClassInfoImpl.PropertyGroup.ELEMENT;
                ++groupCount;
            }
            if (r1 != null || r2 != null || xae != null || mx != null || ov != null) {
                group = ClassInfoImpl.PropertyGroup.ELEMENT_REF;
                ++groupCount;
            }
            if (groupCount > 1) {
                List<Annotation> err = makeSet(new Annotation[] { t, aa, a, v, (Annotation)pickOne(new Annotation[] { e1, e2 }), (Annotation)pickOne(new Annotation[] { r1, r2, xae }) });
                throw new ConflictException(err);
            }
            if (group == null) {
                assert groupCount == 0;
                if (this.nav().isSubClassOf(seed.getRawType(), this.nav().ref(Map.class)) && !seed.hasAnnotation(XmlJavaTypeAdapter.class)) {
                    group = ClassInfoImpl.PropertyGroup.MAP;
                }
                else {
                    group = ClassInfoImpl.PropertyGroup.ELEMENT;
                }
            }
            else if (group.equals(ClassInfoImpl.PropertyGroup.ELEMENT) && this.nav().isSubClassOf(seed.getRawType(), this.nav().ref(Map.class)) && !seed.hasAnnotation(XmlJavaTypeAdapter.class)) {
                group = ClassInfoImpl.PropertyGroup.MAP;
            }
            if ((secondaryAnnotations & group.allowedsecondaryAnnotations) != 0x0) {
                for (ClassInfoImpl.SecondaryAnnotation sa : ClassInfoImpl.SECONDARY_ANNOTATIONS) {
                    if (!group.allows(sa)) {
                        Class[] members = sa.members;
                        int length3 = members.length;
                        int k = 0;
                        while (k < length3) {
                            Class<? extends Annotation> m = members[k];
                            Annotation offender = seed.readAnnotation(m);
                            if (offender != null) {
                                this.builder.reportError(new IllegalAnnotationException(Messages.ANNOTATION_NOT_ALLOWED.format(new Object[] { m.getSimpleName() }), offender));
                                return;
                            }
                            ++k;
                        }
                    }
                }
                assert false;
            }
            switch (group) {
                case TRANSIENT: {
                    return;
                }
                case ANY_ATTRIBUTE: {
                    if (this.attributeWildcard != null) {
                        this.builder.reportError(new IllegalAnnotationException(Messages.TWO_ATTRIBUTE_WILDCARDS.format(new Object[] { this.nav().getClassName(this.getClazz()) }), aa, this.attributeWildcard));
                        return;
                    }
                    this.attributeWildcard = seed;
                    if (this.inheritsAttributeWildcard()) {
                        this.builder.reportError(new IllegalAnnotationException(Messages.SUPER_CLASS_HAS_WILDCARD.format(new Object[0]), aa, this.getInheritedAttributeWildcard()));
                        return;
                    }
                    if (!this.nav().isSubClassOf(seed.getRawType(), this.nav().ref(Map.class))) {
                        this.builder.reportError(new IllegalAnnotationException(Messages.INVALID_ATTRIBUTE_WILDCARD_TYPE.format(new Object[] { this.nav().getTypeName(seed.getRawType()) }), aa, this.getInheritedAttributeWildcard()));
                        return;
                    }
                    return;
                }
                case ATTRIBUTE: {
                    this.properties.add(this.createAttributeProperty(seed));
                    return;
                }
                case VALUE: {
                    this.properties.add(this.createValueProperty(seed));
                    return;
                }
                case ELEMENT: {
                    this.properties.add(this.createElementProperty(seed));
                    return;
                }
                case ELEMENT_REF: {
                    this.properties.add(this.createReferenceProperty(seed));
                    return;
                }
                case MAP: {
                    this.properties.add(this.createMapProperty(seed));
                    return;
                }
                default: {
                    assert false;
                    break;
                }
            }
        }
        catch (ClassInfoImpl.ConflictException x) {
            List<Annotation> err2 = x.annotations;
            this.builder.reportError(new IllegalAnnotationException(Messages.MUTUALLY_EXCLUSIVE_ANNOTATIONS.format(new Object[] { this.nav().getClassName(this.getClazz()) + 35 + seed.getName(), ((Annotation)err2.get(0)).annotationType().getName(), ((Annotation)err2.get(1)).annotationType().getName() }), (Annotation)err2.get(0), (Annotation)err2.get(1)));
        }
        catch (ClassInfoImpl.DuplicateException e3) {
            this.builder.reportError(new IllegalAnnotationException(Messages.DUPLICATE_ANNOTATIONS.format(new Object[] { e3.a1.annotationType().getName() }), e3.a1, e3.a2));
        }
    }
    
    protected ReferencePropertyInfoImpl<T, C, F, M> createReferenceProperty(PropertySeed<T, C, F, M> seed) {
        return new ReferencePropertyInfoImpl(this, seed);
    }
    
    protected AttributePropertyInfoImpl<T, C, F, M> createAttributeProperty(PropertySeed<T, C, F, M> seed) {
        return new AttributePropertyInfoImpl(this, seed);
    }
    
    protected ValuePropertyInfoImpl<T, C, F, M> createValueProperty(PropertySeed<T, C, F, M> seed) {
        return new ValuePropertyInfoImpl(this, seed);
    }
    
    protected ElementPropertyInfoImpl<T, C, F, M> createElementProperty(PropertySeed<T, C, F, M> seed) {
        return new ElementPropertyInfoImpl(this, seed);
    }
    
    protected MapPropertyInfoImpl<T, C, F, M> createMapProperty(PropertySeed<T, C, F, M> seed) {
        return new MapPropertyInfoImpl(this, seed);
    }
    
    private void findGetterSetterProperties(XmlAccessType at) {
        Map<String, M> getters = new LinkedHashMap();
        Map<String, M> setters = new LinkedHashMap();
        C c = this.clazz;
        do {
            this.collectGetterSetters(this.clazz, getters, setters);
            c = this.nav().getSuperClass(c);
        } while (this.shouldRecurseSuperClass(c));
        Set<String> complete = new TreeSet(getters.keySet());
        complete.retainAll(setters.keySet());
        this.resurrect(getters, complete);
        this.resurrect(setters, complete);
        for (String name : complete) {
            M getter = getters.get(name);
            M setter = setters.get(name);
            Annotation[] ga = (getter != null) ? this.reader().getAllMethodAnnotations(getter, new MethodLocatable(this, getter, this.nav())) : ClassInfoImpl.EMPTY_ANNOTATIONS;
            Annotation[] sa = (setter != null) ? this.reader().getAllMethodAnnotations(setter, new MethodLocatable(this, setter, this.nav())) : ClassInfoImpl.EMPTY_ANNOTATIONS;
            boolean hasAnnotation = hasJAXBAnnotation(ga) || hasJAXBAnnotation(sa);
            boolean isOverriding = false;
            if (!hasAnnotation) {
                isOverriding = (getter != null && this.nav().isOverriding(getter, c) && setter != null && this.nav().isOverriding(setter, c));
            }
            if ((at == XmlAccessType.PROPERTY && !isOverriding) || (at == XmlAccessType.PUBLIC_MEMBER && this.isConsideredPublic(getter) && this.isConsideredPublic(setter) && !isOverriding) || hasAnnotation) {
                if (getter != null && setter != null && !this.nav().isSameType(this.nav().getReturnType(getter), this.nav().getMethodParameters(setter)[0])) {
                    this.builder.reportError(new IllegalAnnotationException(Messages.GETTER_SETTER_INCOMPATIBLE_TYPE.format(new Object[] { this.nav().getTypeName(this.nav().getReturnType(getter)), this.nav().getTypeName(this.nav().getMethodParameters(setter)[0]) }), new MethodLocatable(this, getter, this.nav()), new MethodLocatable(this, setter, this.nav())));
                }
                else {
                    Annotation[] r;
                    if (ga.length == 0) {
                        r = sa;
                    }
                    else if (sa.length == 0) {
                        r = ga;
                    }
                    else {
                        r = new Annotation[ga.length + sa.length];
                        System.arraycopy(ga, 0, r, 0, ga.length);
                        System.arraycopy(sa, 0, r, ga.length, sa.length);
                    }
                    this.addProperty(this.createAccessorSeed(getter, setter), r, false);
                }
            }
            else {
                continue;
            }
        }
        getters.keySet().removeAll(complete);
        setters.keySet().removeAll(complete);
    }
    
    private void collectGetterSetters(C c, Map<String, M> getters, Map<String, M> setters) {
        C sc = this.nav().getSuperClass(c);
        if (this.shouldRecurseSuperClass(sc)) {
            this.collectGetterSetters(sc, getters, setters);
        }
        Collection<? extends M> methods = this.nav().getDeclaredMethods(c);
        Map<String, List<M>> allSetters = new LinkedHashMap();
        for (M method : methods) {
            boolean used = false;
            if (this.nav().isBridgeMethod(method)) {
                continue;
            }
            String name = this.nav().getMethodName(method);
            int arity = this.nav().getMethodParameters(method).length;
            if (this.nav().isStaticMethod(method)) {
                this.ensureNoAnnotation(method);
            }
            else {
                String propName = getPropertyNameFromGetMethod(name);
                if (propName != null && arity == 0) {
                    getters.put(propName, method);
                    used = true;
                }
                propName = getPropertyNameFromSetMethod(name);
                if (propName != null && arity == 1) {
                    List<M> propSetters = (List)allSetters.get(propName);
                    if (null == propSetters) {
                        propSetters = new ArrayList();
                        allSetters.put(propName, propSetters);
                    }
                    propSetters.add(method);
                    used = true;
                }
                if (!used) {
                    this.ensureNoAnnotation(method);
                }
                else {
                    continue;
                }
            }
        }
        for (Map.Entry<String, M> entry : getters.entrySet()) {
            String propName2 = (String)entry.getKey();
            M getter = entry.getValue();
            List<M> propSetters2 = (List)allSetters.remove(propName2);
            if (null == propSetters2) {
                continue;
            }
            T getterType = this.nav().getReturnType(getter);
            for (M setter : propSetters2) {
                T setterType = this.nav().getMethodParameters(setter)[0];
                if (this.nav().isSameType(setterType, getterType)) {
                    setters.put(propName2, setter);
                    break;
                }
                continue;
            }
        }
        for (Map.Entry<String, List<M>> e : allSetters.entrySet()) {
            setters.put(e.getKey(), ((List)e.getValue()).get(0));
        }
    }
    
    private boolean shouldRecurseSuperClass(C sc) {
        return sc != null && (this.builder.isReplaced(sc) || this.reader().hasClassAnnotation(sc, XmlTransient.class));
    }
    
    private boolean isConsideredPublic(M m) {
        return m == null || this.nav().isPublicMethod(m);
    }
    
    private void resurrect(Map<String, M> methods, Set<String> complete) {
        for (Map.Entry<String, M> e : methods.entrySet()) {
            if (complete.contains(e.getKey())) {
                continue;
            }
            if (hasJAXBAnnotation(this.reader().getAllMethodAnnotations(e.getValue(), this))) {
                complete.add(e.getKey());
            }
            else {
                continue;
            }
        }
    }
    
    private void ensureNoAnnotation(M method) {
        Annotation[] annotations = this.reader().getAllMethodAnnotations(method, this);
        int length = annotations.length;
        int i = 0;
        while (i < length) {
            Annotation a = annotations[i];
            if (isJAXBAnnotation(a)) {
                this.builder.reportError(new IllegalAnnotationException(Messages.ANNOTATION_ON_WRONG_METHOD.format(new Object[0]), a));
                return;
            }
            ++i;
        }
    }
    
    private static boolean isJAXBAnnotation(Annotation a) {
        return ClassInfoImpl.ANNOTATION_NUMBER_MAP.containsKey(a.annotationType());
    }
    
    private static boolean hasJAXBAnnotation(Annotation[] annotations) {
        return getSomeJAXBAnnotation(annotations) != null;
    }
    
    private static Annotation getSomeJAXBAnnotation(Annotation[] annotations) {
        int length = annotations.length;
        int i = 0;
        while (i < length) {
            Annotation a = annotations[i];
            if (isJAXBAnnotation(a)) {
                return a;
            }
            ++i;
        }
        return null;
    }
    
    private static String getPropertyNameFromGetMethod(String name) {
        if (name.startsWith("get") && name.length() > 3) {
            return name.substring(3);
        }
        if (name.startsWith("is") && name.length() > 2) {
            return name.substring(2);
        }
        return null;
    }
    
    private static String getPropertyNameFromSetMethod(String name) {
        if (name.startsWith("set") && name.length() > 3) {
            return name.substring(3);
        }
        return null;
    }
    
    protected PropertySeed<T, C, F, M> createFieldSeed(F f) {
        return new FieldPropertySeed(this, f);
    }
    
    protected PropertySeed<T, C, F, M> createAccessorSeed(M getter, M setter) {
        return new GetterSetterPropertySeed(this, getter, setter);
    }
    
    public final boolean isElement() {
        return this.elementName != null;
    }
    
    public boolean isAbstract() {
        return this.nav().isAbstract(this.clazz);
    }
    
    public boolean isOrdered() {
        return this.propOrder != null;
    }
    
    public final boolean isFinal() {
        return this.nav().isFinal(this.clazz);
    }
    
    public final boolean hasSubClasses() {
        return this.hasSubClasses;
    }
    
    public final boolean hasAttributeWildcard() {
        return this.declaresAttributeWildcard() || this.inheritsAttributeWildcard();
    }
    
    public final boolean inheritsAttributeWildcard() {
        return this.getInheritedAttributeWildcard() != null;
    }
    
    public final boolean declaresAttributeWildcard() {
        return this.attributeWildcard != null;
    }
    
    private PropertySeed<T, C, F, M> getInheritedAttributeWildcard() {
        ClassInfoImpl<T, C, F, M> c = this.getBaseClass();
        while (c != null) {
            if (c.attributeWildcard != null) {
                return c.attributeWildcard;
            }
            c = c.getBaseClass();
        }
        return null;
    }
    
    public final QName getElementName() {
        return this.elementName;
    }
    
    public final QName getTypeName() {
        return this.typeName;
    }
    
    public final boolean isSimpleType() {
        List<? extends PropertyInfo> props = this.getProperties();
        return props.size() == 1 && ((PropertyInfo)props.get(0)).kind() == PropertyKind.VALUE;
    }
    
    @Override
    void link() {
        this.getProperties();
        Map<String, PropertyInfoImpl> names = new HashMap();
        for (PropertyInfoImpl<T, C, F, M> p : this.properties) {
            p.link();
            PropertyInfoImpl<T, C, F, M> old = (PropertyInfoImpl<T, C, F, M>)names.put(p.getName(), p);
            if (old != null) {
                this.builder.reportError(new IllegalAnnotationException(Messages.PROPERTY_COLLISION.format(new Object[] { p.getName() }), p, old));
            }
            else {
                continue;
            }
        }
        super.link();
    }
    
    public Location getLocation() {
        return this.nav().getClassLocation(this.clazz);
    }
    
    private boolean hasFactoryConstructor(XmlType t) {
        if (t == null) {
            return false;
        }
        String method = t.factoryMethod();
        T fClass = this.reader().getClassValue(t, "factoryClass");
        if (method.length() > 0) {
            if (this.nav().isSameType(fClass, this.nav().ref(XmlType.DEFAULT.class))) {
                fClass = this.nav().use(this.clazz);
            }
            for (M m : this.nav().getDeclaredMethods(this.nav().asDecl(fClass))) {
                if (this.nav().getMethodName(m).equals(method) && this.nav().isSameType(this.nav().getReturnType(m), this.nav().use(this.clazz)) && this.nav().getMethodParameters(m).length == 0 && this.nav().isStaticMethod(m)) {
                    this.factoryMethod = m;
                    break;
                }
                continue;
            }
            if (this.factoryMethod == null) {
                this.builder.reportError(new IllegalAnnotationException(Messages.NO_FACTORY_METHOD.format(new Object[] { this.nav().getClassName(this.nav().asDecl(fClass)), method }), this));
            }
        }
        else if (!this.nav().isSameType(fClass, this.nav().ref(XmlType.DEFAULT.class))) {
            this.builder.reportError(new IllegalAnnotationException(Messages.FACTORY_CLASS_NEEDS_FACTORY_METHOD.format(new Object[] { this.nav().getClassName(this.nav().asDecl(fClass)) }), this));
        }
        return this.factoryMethod != null;
    }
    
    public Method getFactoryMethod() {
        return (Method)this.factoryMethod;
    }
    
    public String toString() {
        return "ClassInfo(" + this.clazz + 41;
    }
    
    static {
        SECONDARY_ANNOTATIONS = ClassInfoImpl.SecondaryAnnotation.values();
        EMPTY_ANNOTATIONS = new Annotation[0];
        ANNOTATION_NUMBER_MAP = new HashMap();
        Class[] annotations = { XmlTransient.class, XmlAnyAttribute.class, XmlAttribute.class, XmlValue.class, XmlElement.class, XmlElements.class, XmlElementRef.class, XmlElementRefs.class, XmlAnyElement.class, XmlMixed.class, OverrideAnnotationOf.class };
        HashMap<Class, Integer> m = ClassInfoImpl.ANNOTATION_NUMBER_MAP;
        for (Class c : annotations) {
            m.put(c, Integer.valueOf(m.size()));
        }
        int index = 20;
        ClassInfoImpl.SecondaryAnnotation[] secondary_ANNOTATIONS = ClassInfoImpl.SECONDARY_ANNOTATIONS;
        for (int length2 = secondary_ANNOTATIONS.length, j = 0; j < length2; ++index, ++j) {
            ClassInfoImpl.SecondaryAnnotation sa = secondary_ANNOTATIONS[j];
            for (Class member : sa.members) {
                m.put(member, Integer.valueOf(index));
            }
        }
        DEFAULT_ORDER = new String[0];
    }
    
    private final class PropertySorter extends HashMap<String, Integer> implements Comparator<PropertyInfoImpl>
    {
        PropertyInfoImpl[] used;
        private Set<String> collidedNames;
        
        PropertySorter() {
            super(ClassInfoImpl.this.propOrder.length);
            this.used = new PropertyInfoImpl[ClassInfoImpl.this.propOrder.length];
            for (String name : ClassInfoImpl.this.propOrder) {
                if (this.put(name, Integer.valueOf(this.size())) != null) {
                    ClassInfoImpl.this.builder.reportError(new IllegalAnnotationException(Messages.DUPLICATE_ENTRY_IN_PROP_ORDER.format(new Object[] { name }), ClassInfoImpl.this));
                }
            }
        }
        
        @Override
        public int compare(PropertyInfoImpl o1, PropertyInfoImpl o2) {
            int lhs = this.checkedGet(o1);
            int rhs = this.checkedGet(o2);
            return lhs - rhs;
        }
        
        private int checkedGet(PropertyInfoImpl p) {
            Integer i = (Integer)this.get(p.getName());
            if (i == null) {
                if (p.kind().isOrdered) {
                    ClassInfoImpl.this.builder.reportError(new IllegalAnnotationException(Messages.PROPERTY_MISSING_FROM_ORDER.format(new Object[] { p.getName() }), p));
                }
                i = Integer.valueOf(this.size());
                this.put(p.getName(), i);
            }
            int ii = (int)i;
            if (ii < this.used.length) {
                if (this.used[ii] != null && this.used[ii] != p) {
                    if (this.collidedNames == null) {
                        this.collidedNames = new HashSet();
                    }
                    if (this.collidedNames.add(p.getName())) {
                        ClassInfoImpl.this.builder.reportError(new IllegalAnnotationException(Messages.DUPLICATE_PROPERTIES.format(new Object[] { p.getName() }), p, this.used[ii]));
                    }
                }
                this.used[ii] = p;
            }
            return (int)i;
        }
        
        public void checkUnusedProperties() {
            for (int i = 0; i < this.used.length; ++i) {
                if (this.used[i] == null) {
                    String unusedName = ClassInfoImpl.this.propOrder[i];
                    String nearest = EditDistance.findNearest(unusedName, new AbstractList<String>() {
                        @Override
                        public String get(int index) {
                            return ((PropertyInfoImpl)ClassInfoImpl.this.properties.get(index)).getName();
                        }
                        
                        @Override
                        public int size() {
                            return ClassInfoImpl.this.properties.size();
                        }
                    });
                    boolean isOverriding = i <= ClassInfoImpl.this.properties.size() - 1 && ((PropertyInfoImpl)ClassInfoImpl.this.properties.get(i)).hasAnnotation(OverrideAnnotationOf.class);
                    if (!isOverriding) {
                        ClassInfoImpl.this.builder.reportError(new IllegalAnnotationException(Messages.PROPERTY_ORDER_CONTAINS_UNUSED_ENTRY.format(new Object[] { unusedName, nearest }), ClassInfoImpl.this));
                    }
                }
            }
        }
        
        static int access$000(ClassInfoImpl.PropertySorter x0, PropertyInfoImpl x1) {
            return x0.checkedGet(x1);
        }
    }
    
    private static final class ConflictException extends Exception
    {
        final List<Annotation> annotations;
        
        public ConflictException(List<Annotation> one) {
            this.annotations = one;
        }
    }
    
    private static final class DuplicateException extends Exception
    {
        final Annotation a1;
        final Annotation a2;
        
        public DuplicateException(Annotation a1, Annotation a2) {
            this.a1 = a1;
            this.a2 = a2;
        }
    }
    
    private enum SecondaryAnnotation
    {
        JAVA_TYPE(1, new Class[] { XmlJavaTypeAdapter.class }), 
        ID_IDREF(2, new Class[] { XmlID.class, XmlIDREF.class }), 
        BINARY(4, new Class[] { XmlInlineBinaryData.class, XmlMimeType.class, XmlAttachmentRef.class }), 
        ELEMENT_WRAPPER(8, new Class[] { XmlElementWrapper.class }), 
        LIST(16, new Class[] { XmlList.class }), 
        SCHEMA_TYPE(32, new Class[] { XmlSchemaType.class });
        
        final int bitMask;
        final Class<? extends Annotation>[] members;
        
        private SecondaryAnnotation(int bitMask, Class<? extends Annotation>[] members) {
            this.bitMask = bitMask;
            this.members = members;
        }
    }
    
    private enum PropertyGroup
    {
        TRANSIENT(new boolean[] { false, false, false, false, false, false }), 
        ANY_ATTRIBUTE(new boolean[] { true, false, false, false, false, false }), 
        ATTRIBUTE(new boolean[] { true, true, true, false, true, true }), 
        VALUE(new boolean[] { true, true, true, false, true, true }), 
        ELEMENT(new boolean[] { true, true, true, true, true, true }), 
        ELEMENT_REF(new boolean[] { true, false, false, true, false, false }), 
        MAP(new boolean[] { false, false, false, true, false, false });
        
        final int allowedsecondaryAnnotations;
        
        private PropertyGroup(boolean[] bits) {
            int mask = 0;
            assert bits.length == ClassInfoImpl.SECONDARY_ANNOTATIONS.length;
            for (int i = 0; i < bits.length; ++i) {
                if (bits[i]) {
                    mask |= ClassInfoImpl.SECONDARY_ANNOTATIONS[i].bitMask;
                }
            }
            this.allowedsecondaryAnnotations = ~mask;
            return;
        }
        
        boolean allows(ClassInfoImpl.SecondaryAnnotation a) {
            return (this.allowedsecondaryAnnotations & a.bitMask) == 0x0;
        }
    }
}
